/*
 *   JVending
 *   Copyright (C) 2004  Shane Isbell
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.jvending.provisioning.stocking.filter;

import java.util.Iterator;
import java.util.List;

import java.util.logging.*;

import org.jvending.provisioning.stocking.par.*;

/**
 * Filter that removes white space from the provisioning.xml file and that removes "/" from all content/icons references
 * within the xml file.
 *
 * @author Shane Isbell
 * @since 1.3a
 */

public final class FormatFilter implements StockingFilter {

    private static Logger logger = Logger.getLogger("FormatFilter");

    public void doFilter(FilterTask filterTask) {
        ProvisioningArchive archive = filterTask.getProvisioningArchive();

        logger.info("Starting the FormatFilter");
        if(archive == null) {
            logger.info("Archive is null. Cannot continue filter.");
            return;
        }

        for(Iterator i = archive.getClientBundle().iterator(); i.hasNext(); ) {
            ClientBundleType bundle = (ClientBundleType) i.next();
            formatClientBundle(bundle);
        }
    }

    public void formatClientBundle(ClientBundleType bundle) {
        bundle.setContentId(bundle.getContentId().trim());

        String version = bundle.getVersion();
        if(version != null) bundle.setVersion(version.trim());

        formatUserDescriptions(bundle.getUserDescriptions());

        DescriptorFileType descriptorFileType = bundle.getDescriptorFile();
        if(descriptorFileType != null) formatDescriptorFile(descriptorFileType);

        ContentFileType contentFileType = bundle.getContentFile();
        if(contentFileType != null) formatContentFile(contentFileType);

        String bundleType = bundle.getBundleType().trim();
        bundle.setBundleType(bundleType);
        
        formatDeviceRequirements(bundle.getDeviceRequirement());

        for(Iterator i = bundle.getPreview().iterator(); i.hasNext(); ) {
            PreviewType type = (PreviewType) i.next();
            formatPreview(type);
        }

        for(Iterator i = bundle.getCopyright().iterator(); i.hasNext(); ) {
            CopyrightType type = (CopyrightType) i.next();
            formatCopyright(type);
        }
    }

    private void formatUserDescriptions(UserDescriptionsType userDescriptionsType) {
        if(userDescriptionsType == null) return;
        for(Iterator i = userDescriptionsType.getIcon().iterator(); i.hasNext(); ) {
            IconType iconType = (IconType) i.next();
            formatIcon(iconType);
        }

        for(Iterator i = userDescriptionsType.getDisplayName().iterator(); i.hasNext(); ) {
            DisplayNameType type = (DisplayNameType) i.next();
            formatDisplayName(type);
        }

        for(Iterator i = userDescriptionsType.getDescription().iterator(); i.hasNext(); ) {
            DescriptionType type = (DescriptionType) i.next();
            formatDescription(type);
        }
    }

    private void formatDeviceRequirements(List deviceRequirements) {
        if(deviceRequirements == null) return;
        for(Iterator i = deviceRequirements.iterator(); i.hasNext(); ) {
            DeviceRequirementType type = (DeviceRequirementType) i.next();
            type.setRequirementName(type.getRequirementName().trim());
            type.setRequirementValue(type.getRequirementValue().trim());
        }
    }

    private void formatPreview(PreviewType type) {
        if(type == null) return;
        String value = type.getValue();
        if(value != null) {
            value = value.trim();
            if(value.startsWith("/")) value = value.substring(1);
            type.setValue(value);
        }
    }

    private void formatIcon(IconType type) {
        if(type == null) return;
        String value = type.getValue();
        if(value != null) {
            value = value.trim();
            if(value.startsWith("/")) value = value.substring(1);
            type.setValue(value);
        }
    }

    private void formatDisplayName(DisplayNameType type) {
        if(type == null) return;
        String value = type.getValue();
        if(value != null) {
            type.setValue(value.trim());
        }
    }

    private void formatDescription(DescriptionType type) {
        if(type == null) return;
        String value = type.getValue();
        if(value != null) {
            type.setValue(value.trim());
        }
    }

    private void formatCopyright(CopyrightType type) {
        if(type == null) return;
        String value = type.getValue();
        if(value != null) {
            type.setValue(value.trim());
        }
    }

    private void formatDescriptorFile(DescriptorFileType type) {
        if(type == null) return;
        String value = type.getValue();
        if(value != null) {
            value = value.trim();
            if(value.startsWith("/")) value = value.substring(1);
            type.setValue(value);
        }
    }

    private void formatContentFile(ContentFileType type) {
        if(type == null) return;
        String value = type.getValue();
        if(value != null) {
            value = value.trim();
            if(value.startsWith("/")) value = value.substring(1);
            type.setValue(value);
        }
    }

}